@@ -15,7 +15,7 @@ use Scope::Guard;
 
 use Exporter qw(import);
 
-our $VERSION = '0.17';
+our $VERSION = '0.19';
 our @EXPORT_OK = qw(start_server restart_server server_ports);
 
 my @signals_received;
@@ -46,7 +46,7 @@ sub start_server {
         if ! ref $paths && defined $paths;
     croak "mandatory option ``exec'' is missing or is not an arrayref\n"
         unless $opts->{exec} && ref $opts->{exec} eq 'ARRAY';
-    
+
     # set envs
     $ENV{ENVDIR} = $opts->{envdir}
         if defined $opts->{envdir};
@@ -56,7 +56,7 @@ sub start_server {
         if defined $opts->{kill_old_delay};
     $ENV{AUTO_RESTART_INTERVAL} = $opts->{auto_restart_interval}
         if defined $opts->{auto_restart_interval};
-    
+
     # open pid file
     my $pid_file_guard = sub {
         return unless $opts->{pid_file};
@@ -153,11 +153,11 @@ sub start_server {
     # setup signal handlers
     $SIG{$_} = sub {
         push @signals_received, $_[0];
-    } for (qw/INT TERM HUP/);
+    } for (qw/INT TERM HUP ALRM/);
     $SIG{PIPE} = 'IGNORE';
     
     # setup status monitor
-    my ($current_worker, %old_workers);
+    my ($current_worker, %old_workers, $last_restart_time);
     my $update_status = $opts->{status_file}
         ? sub {
             my $tmpfn = "$opts->{status_file}.$$";
@@ -176,103 +176,144 @@ sub start_server {
                 or die "failed to rename $tmpfn to $opts->{status_file}:$!";
         } : sub {
         };
-    
+
+    # setup the start_worker function
+    my $start_worker = sub {
+        my $pid;
+        while (1) {
+            $ENV{SERVER_STARTER_GENERATION}++;
+            $pid = fork;
+            die "fork(2) failed:$!"
+                unless defined $pid;
+            if ($pid == 0) {
+                my @args = @{$opts->{exec}};
+                # child process
+                if (defined $opts->{dir}) {
+                    chdir $opts->{dir} or die "failed to chdir:$!";
+                }
+                { exec { $args[0] } @args };
+                print STDERR "failed to exec $args[0]$!";
+                exit(255);
+            }
+            print STDERR "starting new worker $pid\n";
+            sleep $opts->{interval};
+            if ((grep { $_ ne 'HUP' } @signals_received)
+                    || waitpid($pid, WNOHANG) <= 0) {
+                last;
+            }
+            print STDERR "new worker $pid seems to have failed to start, exit status:$?\n";
+        }
+        # ready, update the environment
+        $current_worker = $pid;
+        $last_restart_time = time;
+        $update_status->();
+    };
+
+    # setup the wait function
+    my $wait = sub {
+        my $block = @signals_received == 0;
+        my @r;
+        if ($block && $ENV{ENABLE_AUTO_RESTART}) {
+            alarm(1);
+            @r = wait3($block);
+            alarm(0);
+        } else {
+            @r = wait3($block);
+        }
+        return @r;
+    };
+
+    # setup the cleanup function
+    my $cleanup = sub {
+        my $sig = shift;
+        my $term_signal = $sig eq 'TERM' ? $opts->{signal_on_term} : 'TERM';
+        $old_workers{$current_worker} = $ENV{SERVER_STARTER_GENERATION};
+        undef $current_worker;
+        print STDERR "received $sig, sending $term_signal to all workers:",
+            join(',', sort keys %old_workers), "\n";
+        kill $term_signal, $_
+            for sort keys %old_workers;
+        while (%old_workers) {
+            if (my @r = wait3(1)) {
+                my ($died_worker, $status) = @r;
+                print STDERR "worker $died_worker died, status:$status\n";
+                delete $old_workers{$died_worker};
+                $update_status->();
+            }
+        }
+        print STDERR "exiting\n";
+    };
+
     # the main loop
-    my $term_signal = 0;
-    $current_worker = _start_worker($opts);
-    $update_status->();
-    my $auto_restart_interval = 0;
-    my $last_restart_time = time();
-    my $restart_flag = 0;
+    $start_worker->();
     while (1) {
+        # wait for next signal (or when auto-restart becomes necessary)
+        my @r = $wait->();
+        # reload env if necessary
         my %loaded_env = _reload_env();
         my @loaded_env_keys = keys %loaded_env;
         local @ENV{@loaded_env_keys} = map { $loaded_env{$_} } (@loaded_env_keys);
-        if ($ENV{ENABLE_AUTO_RESTART}) {
-            # restart workers periodically
-            $auto_restart_interval = $ENV{AUTO_RESTART_INTERVAL} ||= 360;
-        }
-        sleep(1);
-        my $died_worker = -1;
-        my $status = -1;
-        while (1) {
-            $died_worker = waitpid(-1, WNOHANG);
-            $status = $?;
-            last if ($died_worker <= 0);
+        $ENV{AUTO_RESTART_INTERVAL} ||= 360
+            if $ENV{ENABLE_AUTO_RESTART};
+        # restart if worker died
+        if (@r) {
+            my ($died_worker, $status) = @r;
             if ($died_worker == $current_worker) {
                 print STDERR "worker $died_worker died unexpectedly with status:$status, restarting\n";
-                $current_worker = _start_worker($opts);
-                $last_restart_time = time();
+                $start_worker->();
             } else {
                 print STDERR "old worker $died_worker died, status:$status\n";
                 delete $old_workers{$died_worker};
-                # don't update the status file if restart is scheduled and died_worker is the last one
-                if ($restart_flag == 0 || scalar(keys %old_workers) != 0) {
-                    $update_status->();
-                }
+                $update_status->();
             }
         }
-        if ($auto_restart_interval > 0 && scalar(@signals_received) == 0 &&
-            time() > $last_restart_time + $auto_restart_interval) {
-            print STDERR "autorestart triggered (interval=$auto_restart_interval)\n";
-            $restart_flag = 1;
-            if (time() > $last_restart_time + $auto_restart_interval * 2) {
-                print STDERR "force autorestart triggered\n";
-                $restart_flag = 2;
-            }
-        }
-        my $num_old_workers = scalar(keys %old_workers);
-        for (; @signals_received; shift @signals_received) {
-            if ($signals_received[0] eq 'HUP') {
-                print STDERR "received HUP (num_old_workers=$num_old_workers)\n";
-                $restart_flag = 1;
+        # handle signals
+        my $restart;
+        while (@signals_received) {
+            my $sig = shift @signals_received;
+            if ($sig eq 'HUP') {
+                print STDERR "received HUP, spawning a new worker\n";
+                $restart = 1;
+                last;
+            } elsif ($sig eq 'ALRM') {
+                # skip
             } else {
-                $term_signal = $signals_received[0] eq 'TERM' ? $opts->{signal_on_term} : 'TERM';
-                goto CLEANUP;
+                return $cleanup->($sig);
             }
         }
-        if ($restart_flag > 1 || ($restart_flag > 0 && $num_old_workers == 0)) {
-            print STDERR "spawning a new worker (num_old_workers=$num_old_workers)\n";
+        if (! $restart && $ENV{ENABLE_AUTO_RESTART}) {
+            my $auto_restart_interval = $ENV{AUTO_RESTART_INTERVAL};
+            my $elapsed_since_restart = time - $last_restart_time;
+            if ($elapsed_since_restart >= $auto_restart_interval && ! %old_workers) {
+                print STDERR "autorestart triggered (interval=$auto_restart_interval)\n";
+                $restart = 1;
+            } elsif ($elapsed_since_restart >= $auto_restart_interval * 2) {
+                print STDERR "autorestart triggered (forced, interval=$auto_restart_interval)\n";
+                $restart = 1;
+            }
+        }
+        # restart if requested
+        if ($restart) {
             $old_workers{$current_worker} = $ENV{SERVER_STARTER_GENERATION};
-            $current_worker = _start_worker($opts);
-            $last_restart_time = time();
-            $restart_flag = 0;
-            $update_status->();
+            $start_worker->();
             print STDERR "new worker is now running, sending $opts->{signal_on_hup} to old workers:";
             if (%old_workers) {
                 print STDERR join(',', sort keys %old_workers), "\n";
             } else {
                 print STDERR "none\n";
             }
-            my $kill_old_delay = $ENV{KILL_OLD_DELAY} || 0;
-            $kill_old_delay ||= 5 if $ENV{ENABLE_AUTO_RESTART};
-            print STDERR "sleep $kill_old_delay secs\n";
-            sleep($kill_old_delay) if $kill_old_delay > 0;
+            my $kill_old_delay = defined $ENV{KILL_OLD_DELAY} ? $ENV{KILL_OLD_DELAY} : $ENV{ENABLE_AUTO_RESTART} ? 5 : 0;
+            if ($kill_old_delay != 0) {
+                print STDERR "sleeping $kill_old_delay secs before killing old workers\n";
+                sleep $kill_old_delay;
+            }
             print STDERR "killing old workers\n";
             kill $opts->{signal_on_hup}, $_
                 for sort keys %old_workers;
         }
     }
-    
- CLEANUP:
-    # cleanup
-    $old_workers{$current_worker} = $ENV{SERVER_STARTER_GENERATION};
-    undef $current_worker;
 
-    print STDERR "received $signals_received[0], sending $term_signal to all workers:",
-        join(',', sort keys %old_workers), "\n";
-    kill $term_signal, $_
-        for sort keys %old_workers;
-    while (%old_workers) {
-        if (my @r = wait3(1)) {
-            my ($died_worker, $status) = @r;
-            print STDERR "worker $died_worker died, status:$status\n";
-            delete $old_workers{$died_worker};
-            $update_status->();
-        }
-    }
-    
-    print STDERR "exiting\n";
+    die "unreachable";
 }
 
 sub restart_server {
@@ -312,7 +353,7 @@ sub restart_server {
     # wait for the generation
     while (1) {
         my @gens = $get_generations->();
-        last if scalar(@gens) == 1 && $gens[0] >= $wait_for;
+        last if scalar(@gens) == 1 && $gens[0] == $wait_for;
         sleep 1;
     }
 }
@@ -341,35 +382,6 @@ sub _reload_env {
     return %env;
 }
 
-sub _start_worker {
-    my $opts = shift;
-    my $pid;
-    while (1) {
-        $ENV{SERVER_STARTER_GENERATION}++;
-        $pid = fork;
-        die "fork(2) failed:$!"
-            unless defined $pid;
-        if ($pid == 0) {
-            my @args = @{$opts->{exec}};
-            # child process
-            if (defined $opts->{dir}) {
-                chdir $opts->{dir} or die "failed to chdir:$!";
-            }
-            { exec { $args[0] } @args };
-            print STDERR "failed to exec $args[0]$!";
-            exit(255);
-        }
-        print STDERR "starting new worker $pid\n";
-        sleep $opts->{interval};
-        if ((grep { $_ ne 'HUP' } @signals_received)
-                || waitpid($pid, WNOHANG) <= 0) {
-            last;
-        }
-        print STDERR "new worker $pid seems to have failed to start, exit status:$?\n";
-    }
-    $pid;
-}
-
 1;
 __END__
 
